{-# LANGUAGE GADTs                 #-}
{-# LANGUAGE DataKinds             #-}
{-# LANGUAGE TypeFamilies          #-}
{-# LANGUAGE TypeOperators         #-}
{-# LANGUAGE StandaloneDeriving    #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# LANGUAGE NoImplicitPrelude     #-}
{-# LANGUAGE UndecidableInstances  #-}
{-# LANGUAGE TemplateHaskell       #-}
{-# LANGUAGE QuasiQuotes           #-}
{-# LANGUAGE PolyKinds             #-}
{-# LANGUAGE UnicodeSyntax         #-}

-- | Defines natural numbers on both term and type level
module Data.Nat where

import Relation.Equality
import Data.Bool
import Data.Integer              (Integer)
import Data.Function             (id, const, (∘))
import Data.TypeClass            (Eq, Show, Num, Enum, Bounded)
import Data.Singletons.TH hiding ((:<), (:>), (%:<), Refl)

import qualified Data.TypeClass as T

singletons [d|

  -- | Natural numbers type
  data ℕ ∷ ★ where
    Z ∷ ℕ     -- ^ Zero
    S ∷ ℕ → ℕ -- ^ Successor
    deriving (Eq, Show)

  infixl 6 +
  infixl 7 ×
  
  -- | Predecessor of n  
  pred ∷ ℕ → ℕ
  pred Z     = Z
  pred (S n) = n

  -- | Addition
  (+) ∷ ℕ → ℕ → ℕ
  Z     + m = m
  (S n) + m = S (n + m)

  -- | Subtraction
  (-) ∷ ℕ → ℕ → ℕ
  Z     -    _  = Z
  n     -    Z  = n
  (S n) - (S m) = n - m

  -- | Multiplication
  (×) ∷ ℕ → ℕ → ℕ
  Z     × m = Z
  (S n) × m = (n × m) + m

  -- | Ordering relation <
  (<) ∷ ℕ → ℕ → 𝔹
  Z     < Z     = F
  Z     < (S _) = T
  (S _) < Z     = F
  (S n) < (S m) = n < m

  -- | Ordering relation >
  (>) ∷ ℕ → ℕ → 𝔹
  n > m = (¬) (n < m)

  |]


instance Num ℕ where
  (+) = (+)
  (-) = (-)
  (*) = (×)
  abs = id
  signum = const 1
  negate = id
  fromInteger = fromInteger

instance Enum ℕ where
  toEnum   = fromInteger ∘ T.toInteger
  fromEnum = T.fromInteger ∘ toInteger

instance Bounded ℕ where
  minBound = Z
  maxBound = (∞)


-- | Convert a natural into an 'Integer'
toInteger ∷ ℕ → Integer
toInteger Z     = 0
toInteger (S n) = 1 T.+ toInteger n


-- | Convert an 'Integer' into a natural
fromInteger ∷ Integer → ℕ
fromInteger 0 = Z
fromInteger n
  | n T.<= 0 = 0
  | n T.>  0 = S (fromInteger (T.pred n)) 


-- | Equality for natural numbers
(≡) ∷ ℕ → ℕ → 𝔹 
Z     ≡ Z     = T
(S n) ≡ (S m) = n ≡ m
_     ≡ _     = F


-- | Infinity
(∞) ∷ ℕ
(∞) = S (∞)


-- Proofs

-- | Lemma
switch ∷ Sℕ n → Sℕ m → (n :+ S m) :≡ (S n :+ m)
switch SZ     m = Refl
switch (SS n) m = cong SS (switch n m)

-- | Lemma
succ ∷ Sℕ n → Sℕ m → S (n :+ m) :≡ (n :+ S m)
succ SZ     m = Refl
succ (SS n) m = cong SS (succ n m)

-- | Relation between successor and addition
succ_plus ∷ Sℕ n → (n :+ S Z) :≡ S n
succ_plus n = switch n SZ ∾ plus_right_id (SS n)

-- | Right identity of addition
plus_right_id ∷ Sℕ n → (n :+ Z) :≡ n
plus_right_id SZ     = Refl
plus_right_id (SS n) = cong SS (plus_right_id n)

-- | Associative property of addition
plus_assoc ∷ Sℕ a → Sℕ b → Sℕ c → (a :+ b) :+ c :≡ a :+ (b :+ c)
plus_assoc SZ     _ _ = Refl
plus_assoc (SS a) b c = cong SS (plus_assoc a b c)

-- | Commutative property of addition
plus_commut ∷ Sℕ n → Sℕ m → (n :+ m) :≡ (m :+ n)
plus_commut n   SZ    = plus_right_id n
plus_commut n  (SS m) =
  (sym (succ n m) ∾ cong SS (plus_commut n m)) ∾
  (succ m n ∾ switch m n)
