{-# LANGUAGE GADTs                 #-}
{-# LANGUAGE DataKinds             #-}
{-# LANGUAGE TypeFamilies          #-}
{-# LANGUAGE TypeOperators         #-}
{-# LANGUAGE StandaloneDeriving    #-}
{-# LANGUAGE FlexibleInstances     #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# LANGUAGE NoImplicitPrelude     #-}
{-# LANGUAGE UndecidableInstances  #-}
{-# LANGUAGE TemplateHaskell       #-}
{-# LANGUAGE QuasiQuotes           #-}
{-# LANGUAGE PolyKinds             #-}
{-# LANGUAGE UnicodeSyntax         #-}

-- | Defines a vector type, analogous to a list, that achieves
-- type-safe operation like resizing and concatenation by using
-- new dependantly typed language features in haskell
module Data.Vec where

import Relation.Equality (gcastWith, cong, sym)
import Data.Nat
import Data.Bool
import Data.Char         (Char)
import Data.Function     ((∘))
import Data.TypeClass    (Show, IsList, IsString, Eq) 
import Data.Singletons   (SingI, sing)

import qualified Data.TypeClass as T


infixr 5 :-

-- | The 'Vec' datatype
data Vec (a ∷ ★) ∷ ℕ → ★ where
  Nil  ∷ Vec a Z                   -- ^ empty 'Vec', length 0
  (:-) ∷ a → Vec a n → Vec a (S n) -- ^ "cons" operator


-- | 'String' type alias for vector of characters
type String n = Vec Char n


deriving instance Eq a ⇒ Eq (Vec a n)

instance Show a ⇒ Show (Vec a n) where
  showsPrec d = T.showsPrec d ∘ toList

instance SingI n ⇒ IsList (Vec a n) where
  type Item (Vec a n) = a
  fromList = fromList
  toList   = toList

instance SingI n ⇒ IsString (String n) where
  fromString = fromList



-- * Conversions

-- | Convert a 'Vec' into a List
toList ∷ Vec a n → [a]
toList Nil = []
toList (x :- xs) = x : toList xs


-- | Convert a List into a 'Vec'
--
-- It's not possible to convert an infinite list to a vector
-- as haskell does not permit constructing infinite type,
-- which the resulting vector length would be.
fromList ∷ SingI n ⇒ [a] → Vec a n
fromList = f sing
  where  f ∷ Sℕ n → [a] → Vec a n
         f SZ     _      = Nil
         f (SS n) (x:xs) = x :- f n xs



-- * Basic functions

-- | Vector concatenation
(⧺) ∷ Vec a n → Vec a m → Vec a (n :+ m)
(⧺) (x :- xs) ys = x :- xs ⧺ ys
(⧺) Nil       ys = ys


-- | Extracts the 'head' (ie the first element) of a nonempty 'Vec'.
head ∷ Vec a (S n) → a
head (x :- _) = x


-- | Extracts the last element of a nonempty 'Vec'.
last ∷ Vec a (S n) → a
last (x :- Nil)     = x
last (x :- y :- ys) = last (y :- ys)


-- | Applied to a nonempty 'Vec' returns everything but its first element.
-- > x ≡ head x ⧺ tail x
tail ∷ Vec a (S n) → Vec a n
tail (_ :- xs) = xs


-- | Applied to a nonempty 'Vec' returns everything but its last element.
-- > x ≡ init x ⧺ last x
init ∷ Vec a (S n) → Vec a n
init (x :- Nil)     = Nil
init (x :- y :- ys) = x :- init (y :- ys)


-- | Given a nonempty 'Vec' returns the first element
-- and the rest of the vector in a tuple.
-- > uncons x ≡ (head x, tail x)
uncons ∷ Vec a (S n) → (a, Vec a n)
uncons (x :- xs) = (x, xs)


-- | Test whether a 'Vec' has zero length.
null ∷ Vec a n → 𝔹
null Nil = T
null _   = F


-- | Returns the length (numbers of elements) of a 'Vec'.
length ∷ Vec a n → ℕ
length Nil       = Z
length (_ :- xs) = S (length xs)

-- | Same as 'length' but produces a singleton type 'Sℕ'.
sLength ∷ Vec a n → Sℕ n
sLength Nil       = SZ
sLength (x :- xs) = SS (sLength xs)



-- * Transformations

-- | Applies a function on every element of a 'Vec'.
map ∷ (a → b) → Vec a n → Vec b n
map _ Nil       = Nil
map f (x :- xs) = f x :- map f xs


-- | Reverse the order of the elements of a 'Vec'
-- > reverse ∘ reverse = id
reverse ∷ Vec a n → Vec a n
reverse Nil       = Nil
reverse (y :- ys) = gcastWith proof (reverse ys ⧺ (y :- Nil))
  where proof = succ_plus (sLength ys)


-- | 'intersperse' @s@ takes a 'Vec' and produces one where
-- @s@ is interspersed (ie inserted between) every two elements
-- of the vector
intersperse ∷ a → Vec a n → Vec a (n :+ Pred n)
intersperse _ Nil        = Nil
intersperse _ (x :- Nil) = x :- Nil
intersperse s (x :- xs)  = x :- s :- intersperse s xs


-- | 'foldl' applied to a binary operator, a starting value
-- and a 'Vec' reduces the vector to a single value obtained
-- by sequentially applying the operation from the left to the right.
foldl ∷ (a → b → a) → a → Vec b n → a
foldl _ x Nil       = x
foldl f x (y :- ys) = foldl f (f x y) ys


-- | Same as "fold" but reduces the vector in the opposite
-- direction: from the right to the left.
foldr ∷ (a → b → b) → b → Vec a n → b
foldr f x Nil     = x
foldr f x (y:-ys) = f y (foldr f x ys)


-- | Variant of 'foldl' which requires no starting value but
-- applies only on nonempty 'Vec'
foldl₁ ∷ (a → a → a) → Vec a (S n) → a
foldl₁ f (x :- Nil)     = x
foldl₁ f (x :- y :- ys) = foldl₁ f (f x y :- ys)


-- | As for 'foldl', a variant of 'foldr' with no starting point
foldr₁ ∷ (a → a → a) → Vec a (S n) → a
foldr₁ f (x :- Nil)     = x
foldr₁ f (x :- y :- ys) = f x (foldr₁ f (y :- ys))


-- | Applied to a singleton natural @n@ and a value produces
-- a vector obtained by duplicating the value @n@ times
replicate ∷ Sℕ n → a → Vec a n
replicate SZ     _ = Nil
replicate (SS n) a = a :- replicate n a


-- | 'take' @n@ applied to a 'Vec', returns the first @n@ element
-- of the vector
take ∷ Sℕ n → Vec a m → Vec a n
take SZ      _        = Nil
take (SS n) (x :- xs) = x :- take n xs


-- | 'drop' @n@ applied to a 'Vec', returns every element of
-- the vector but the first @n@
drop ∷ Sℕ n → Vec a m → Vec a (m :- n)
drop SZ     x         = x
drop (SS n) (x :- xs) = drop n xs


